from datetime import datetime, timezone
from flask_securelogin.exception import AppException

from flask_securelogin.models import User
from flask_securelogin.sms.SMSFactory import SMSFactory

import uuid


class AccountManager():
    def __init__(self, db, config=None):
        self._db = db
        self._config = config

    def create_user_w_password(self, username, email, password):
        # check whether the user already exists
        user = User.query.filter_by(email=email).first()
        if user is not None and user.status != 'closed':
            raise AppException("AccountCreationFailure", "AccountAlreadyExist",
                               "failed to create an account as the email already exists")

        # create a new user
        new_user = User(public_id=str(uuid.uuid4()), username=username, auth_type='PASSWORD')
        new_user.email = email
        new_user.set_password(password)

        new_user.status = 'active'
        new_user.time_created = int(datetime.now(timezone.utc).timestamp())

        self._db.session.add(new_user)
        self._db.session.commit()

        return new_user

    def create_user_w_phone(self, username, phone):
        # check whether the user already exists
        user = User.query.filter_by(phone=phone).first()
        if user is not None and user.status != 'closed':
            raise AppException("AccountCreationFailure", "AccountAlreadyExist",
                               "failed to create an account as the phone already exists")

        # create a new user
        new_user = User(public_id=str(uuid.uuid4()), username=username, auth_type='PHONE')
        new_user.phone = phone

        new_user.status = 'active'
        new_user.time_created = int(datetime.now(timezone.utc).timestamp())

        self._db.session.add(new_user)
        self._db.session.commit()

        return new_user

    def account_sanity_check(self, user, auth_type):
        if user is None:
            raise AppException("LoginFailure", "AccountNotExist", "the user doesn't exist")

        if user.auth_type != auth_type:
            raise AppException("LoginFailure", "AccountInvalidAuthType", "authentication method doesn't match")

        if user.status != 'active':
            raise AppException("LoginFailure", "AccountInactive", "the user account is not active anymore")

        return

    def login_by_password(self, email, password):
        user = User.query.filter_by(email=email).first()
        self.account_sanity_check(user, 'PASSWORD')

        if not user.check_password(password):
            raise AppException("LoginFailure", "AccountInvalidPassword", "wrong password")

        return user

    def login_by_phone(self, phone):
        user = User.query.filter_by(phone=phone).first()
        self.account_sanity_check(user, 'PHONE')

        sms = SMSFactory.create_instance(self._db, self._config, phone)
        if not sms.send_sms(phone):
            raise AppException("LoginFailure", "SMSSendFailure", "can't send sms code due to system failure")

        return user

    def login_verify_sms(self, phone, token):
        user = User.query.filter_by(phone=phone).first()
        self.account_sanity_check(user, 'PHONE')

        sms = SMSFactory.create_instance(self._db, self._config, phone)
        if not sms.verify_sms(phone, token):
            raise AppException("LoginFailure", "SMSVerifyFailure", "incorrect sms code")

        return user

    def change_password(self, userid, oldpassword, newpassword):
        user = User.query.filter_by(id=userid).first()
        self.account_sanity_check(user, user.auth_type)

        if not user.check_password(oldpassword):
            raise AppException("PasswordFailure", "ChangePasswordFailed", "password is not correct")

        user.set_password(newpassword)

        self._db.session.commit()



from flask_securelogin.exception import AppException
from flask_securelogin.models import UserActiveToken
from flask_securelogin import secure_auth

from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    decode_token,
)

import uuid


class TokenManager():
    def __init__(self, db):
        self._db = db

    def start_new_session(self, userid):
        # internal session id
        sid = str(uuid.uuid4())

        (access_token, refresh_token) = self.create_tokens(userid, sid, True)
        self.insert_refresh_token_in_db(userid, refresh_token)

        return (access_token, refresh_token)

    def create_tokens(self, userid, sid, fresh):
        # create access_token and refresh_token
        access_token = create_access_token(identity=userid, fresh=fresh)
        refresh_token = create_refresh_token(identity=userid, additional_claims={"sid": sid})

        # update it to database
        return (access_token, refresh_token)

    def insert_refresh_token_in_db(self, userid, refresh_token):
        # decode refresh token
        new_jwt = decode_token(refresh_token)

        # create new record for UserActiveToken
        active_token = UserActiveToken(userid=userid)

        active_token.sessionid = new_jwt['sid']
        active_token.jti = new_jwt['jti']
        active_token.created_at = new_jwt['iat']
        active_token.expired_at = new_jwt['exp']
        active_token.active = True

        # insert to DB
        self._db.session.add(active_token)
        self._db.session.commit()

        return

    def update_refresh_token_in_db(self, active_token, refresh_token):
        # decode refresh token
        new_jwt = decode_token(refresh_token)

        # update the fields
        active_token.jti = new_jwt['jti']
        active_token.created_at = new_jwt['iat']
        active_token.expired_at = new_jwt['exp']
        active_token.active = True

        # update it to DB
        self._db.session.commit()

        return

    def refresh_token_sanity_check(self, active_token, jti, userid):

        if active_token is None:
            raise AppException("LoginFailure", "AccountNoLogin", "The user didn't login")

        if active_token.jti != jti:
            raise AppException("LoginFailure", "AccountObsoleteToken", "The refresh token is already obsolete")

        if active_token.userid != userid:
            raise AppException("LoginFailure", "AccountInvalidSession", "Corrupted refresh_token, userid doesn't match")

        if not active_token.active:
            raise AppException("LoginFailure", "AccountInactiveToken",
                               "The refresh token is inactive, user already logout")

        return

    def refresh_tokens(self, refresh_token_jwt):
        userid = refresh_token_jwt['sub']
        jti = refresh_token_jwt['jti']
        sid = refresh_token_jwt['sid']

        # validate whether the refresh_token is consistent with the active token for the user
        active_token = UserActiveToken.query.filter_by(sessionid=sid).first()

        self.refresh_token_sanity_check(active_token, jti, userid)

        # create new tokens
        (access_token, refresh_token) = self.create_tokens(userid, sid, False)

        # update it it the DB
        self.update_refresh_token_in_db(active_token, refresh_token)

        return (access_token, refresh_token)

    def logout_token(self, refresh_token, access_token_jwt):
        # decode refresh_token
        refresh_token_jwt = decode_token(refresh_token)

        userid = refresh_token_jwt['sub']
        jti = refresh_token_jwt['jti']
        sid = refresh_token_jwt['sid']

        if access_token_jwt is not None and access_token_jwt['sub'] != userid:
            raise AppException("AccountLogoutFailure", "InvalidUserId",
                               "The refresh token is inconsistent with access_token.")

        # validate whether the refresh_token is consistent with the active token for the user
        active_token = UserActiveToken.query.filter_by(sessionid=sid).first()
        self.refresh_token_sanity_check(active_token, jti, userid)

        active_token.active = False
        self._db.session.commit()

    def check_if_token_revoked(self, token_jwt):
        token_type = token_jwt['type']
        if token_type != 'refresh':
            handler = secure_auth.get_check_if_access_token_revoked_handler()
            if handler is not None:
                jti = token_jwt['jti']
                return handler(jti)
            return False

        jti = token_jwt['jti']
        sid = token_jwt['sid']

        active_token = UserActiveToken.query.filter_by(sessionid=sid).first()
        if active_token is None:
            return True

        if active_token.jti != jti:
            return True

        return not active_token.active

    def logout_all_other_sessions(self, access_token_jwt, refresh_token):
        # decode refresh_token
        refresh_token_jwt = decode_token(refresh_token)

        userid = refresh_token_jwt['sub']
        sid = refresh_token_jwt['sid']

        if access_token_jwt['sub'] != userid:
            raise AppException("KickoutAllSessionFailure", "InvalidUserId",
                               "The refresh token is inconsistent with access_token.")

        user_sessions = UserActiveToken.query.filter_by(userid=userid).all()

        for active_token in user_sessions:
            if active_token.sessionid == sid:
                continue

            if not active_token.active:
                continue

            active_token.active = False

        self._db.session.commit()

        return



from flask import jsonify
from werkzeug.http import HTTP_STATUS_CODES


def error_response(status_code, error, **kwargs):
    payload = {'error': error, 'http_code': HTTP_STATUS_CODES.get(status_code, 'Unknown error')}
    for key, value in kwargs.items():
        if key == 'exception':
            ex = value
            payload[key] = {'exception': ex.__class__.__name__, 'message': str(ex)}
        else:
            payload[key] = value

    response = jsonify(payload)
    response.status_code = status_code
    return response


def bad_request(error, message, exception=None):
    return error_response(400, error, message=message, exception=exception)



from flask_securelogin import secure_auth

db = secure_auth.db


class AuthOTP(db.Model):
    alias = db.Column(db.String(120), primary_key=True)
    sms = db.Column(db.String(10))
    created_ts = db.Column(db.BigInteger)



from flask_securelogin.sms.SMSBaseSender import SMSBaseSender
import requests
import json


class UniSMSClient(SMSBaseSender):
    def __init__(self, db, config):
        super().__init__(db, config)
        self.accessKeyId = config['UNISMS_ACCESSKEY_ID']
        self.signature = config['UNISMS_SIGNATURE']

    def umisms_api_call(self, phone, token):
        url = "https://uni.apistd.com/?action=sms.message.send&accessKeyId=" + self.accessKeyId

        headers = {
            "Content-Type": "application/json"
        }

        data = {
            "to": phone,
            "signature": self.signature,
            "templateId": "pub_verif_login",
            "templateData": {"code": token, "ttl": self.ttl}
        }

        response = requests.post(url, headers=headers, data=json.dumps(data))
        return response.text

    def vendor_send_otp(self, phone, token):
        # send token through Uni-SMS api
        resp = self.umisms_api_call(phone, token)

        # parse the response
        x = json.loads(resp)
        code = x['code']
        message = x['message']

        if int(code) != 0:
            self.errors = "Failed to send SMS uni-sms due to " + message + ", error code: " + code
            return False

        return True



from flask_securelogin.sms.SMSBaseSender import SMSBaseSender

import json
from tencentcloud.common import credential
from tencentcloud.common.profile.client_profile import ClientProfile
from tencentcloud.common.profile.http_profile import HttpProfile
from tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException
from tencentcloud.sms.v20210111 import sms_client, models


class TencentAPIClient(SMSBaseSender):
    def __init__(self, db, config):
        super().__init__(db, config)
        self.secretId = config['TENCENT_SECRET_ID']
        self.secretKey = config['TENCENT_SECRET_KEY']
        self.sdkAppId = config['TENCENT_SDK_APPID']
        self.templateId = config['TENCENT_TEMPLATE_ID']
        self.senderId = config['TENCENT_SENDER_ID']
        self.appName = config['TENCENT_APP_NAME']

    def tencent_api_call(self, phone, token):
        # the secretId and secretKey can be acquired the tencent website
        # from https://console.cloud.tencent.com/cam/capi
        # they are configured in the application previously
        cred = credential.Credential(self.secretId, self.secretKey)

        # initiate http profile
        httpProfile = HttpProfile()
        httpProfile.endpoint = "sms.tencentcloudapi.com"

        # initiate an instance of http client
        clientProfile = ClientProfile()
        clientProfile.httpProfile = httpProfile

        client = sms_client.SmsClient(cred, "ap-guangzhou", clientProfile)

        # initiate a request
        req = models.SendSmsRequest()
        params = {
            "PhoneNumberSet": [phone],
            "SmsSdkAppId": self.sdkAppId,
            "TemplateId": self.templateId,
            "TemplateParamSet": [self.appName, token, str(self.ttl)],
            "SenderId": self.senderId
        }
        req.from_json_string(json.dumps(params))

        # send the request to Tencent cloud
        resp = client.SendSms(req)

        # parse the response
        # x = resp.to_json_string()

        return resp

    def vendor_send_otp(self, phone, token):
        # send token through Tencent api
        try:
            resp = self.tencent_api_call(phone, token)

            stat = resp.SendStatusSet[0]
            if stat.Code != 'Ok':
                self.errors = stat.Code + ": " + stat.Message
                return False
        except TencentCloudSDKException as err:
            self.errors = err.code + ": " + err.message
            return False

        return True



from abc import ABCMeta, abstractmethod


class SMSCall:
    __metaclass__ = ABCMeta

    @abstractmethod
    def send_sms(self, phone) -> bool:
        pass

    @abstractmethod
    def verify_sms(self, phone) -> bool:
        pass

    @abstractmethod
    def get_errors(self) -> Exception:
        pass



from werkzeug.security import generate_password_hash, check_password_hash
from flask_securelogin import secure_auth

db = secure_auth.db


class User(db.Model):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    public_id = db.Column(db.String(36), index=True, unique=True)

    username = db.Column(db.String(50), index=True)
    auth_type = db.Column(db.String(10))
    email = db.Column(db.String(120), index=True)
    phone = db.Column(db.String(20), index=True)
    password_hash = db.Column(db.String(128))

    status = db.Column(db.String(10))
    time_created = db.Column(db.Integer)

    def __repr__(self):
        return '<User {}>'.format(self.username)

    def set_password(self, password):
        self.password_hash = generate_password_hash(password)

    def check_password(self, password):
        return check_password_hash(self.password_hash, password)


# The current active refresh_token for a member. The table should be updated
# when every refresh_token is refreshed and set to false when a user logout
class UserActiveToken(db.Model):
    __tablename__ = 'active_tokens'

    # internal id
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)

    # internal session id
    sessionid = db.Column(db.String(36), index=True, unique=True)

    # the user id linked to User table
    userid = db.Column(db.Integer, db.ForeignKey('users.id'), index=True)

    # refresh token
    jti = db.Column(db.String(36), nullable=False, index=True)

    # set to FALSE when a user log out
    active = db.Column(db.Boolean, unique=False, default=True)

    # indicate when the jti was created
    created_at = db.Column(db.Integer, nullable=False)

    # indicate the token expiration time, used to purge the DB
    # for rows that token is expired
    expired_at = db.Column(db.Integer, nullable=False)

    def __repr__(self):
        return '<UserActiveToken {}>'.format(self.sessionid)



from flask_securelogin.sms.models import AuthOTP

import math
import random
import time


class OTPGenerator():
    def __init__(self, db, config):
        self._config = config
        self._db = db
        self._expiration = config['OTP_EXPIRATION']
        self._num_digits = config['OTP_DIGITS']

    def update_token_to_db(self, phone, token):
        otp = AuthOTP.query.filter_by(alias=phone).first()
        if otp is None:
            otp = AuthOTP()
            otp.alias = phone
            otp.created_ts = int(time.time())
            otp.sms = token

            self._db.session.add(otp)
        else:
            otp.created_ts = int(time.time())
            otp.sms = token

        self._db.session.commit()

        return token

    def get_token_from_db(self, phone):
        otp = AuthOTP.query.filter_by(alias=phone).first()
        if otp is None:
            return (None, None)

        return (otp.sms, otp.created_ts)

    def generate_otp(self, phone):
        digits = "0123456789"
        otp = ''
        for i in range(self._num_digits):
            otp += digits[math.floor(random.random() * 10)]

        return otp

    def gen_persist_otp(self, phone):
        otp = self.generate_otp(phone)
        self.update_token_to_db(phone, otp)
        return otp

    def verify_otp(self, phone, token):
        (otp, created_ts) = self.get_token_from_db(phone)
        if otp is None or created_ts is None:
            return False

        if (time.time() - created_ts) > self._expiration * 60:
            return False

        return otp == token



from flask_securelogin.sms.TwilioClient import TwilioClient
from flask_securelogin.sms.UniSMSClient import UniSMSClient
from flask_securelogin.sms.TencentAPIClient import TencentAPIClient
from flask_securelogin import secure_auth


class SMSFactory():
    def get_vendor(config, phone):
        vendor = config['SMS_VENDOR']       # default setting

        # SMS vendor routing
        if 'TWILIO_TESTING_PHONES' in config and phone in config['TWILIO_TESTING_PHONES']:
            vendor = 'twillio'
        elif phone.find('+86') == 0: 		# china phone number
            vendor = 'uni-sms'
        elif phone.find('+1') == 0:			# US phone number
            vendor = 'tencent'

        return vendor.lower()

    def create_instance(db, config, phone):
        handler = secure_auth.get_create_sms_service_handler()
        if handler is None:
            return SMSFactory.create_default_service(db, config, phone)
        else:
            return handler(db, config, phone)

    def create_default_service(db, config, phone):
        vendor = SMSFactory.get_vendor(config, phone)

        if vendor == 'twillio':
            return TwilioClient(db, config)
        elif vendor == 'uni-sms':
            return UniSMSClient(db, config)
        elif vendor == 'tencent':
            return TencentAPIClient(db, config)

        return None



from abc import abstractmethod
from flask_securelogin.sms.SMSCall import SMSCall
from flask_securelogin.sms.OTPGenerator import OTPGenerator


class SMSBaseSender(SMSCall):
    def __init__(self, db, config):
        self.db = db
        self.errors = None
        self.otp_generator = OTPGenerator(db, config)
        self.ttl = config['OTP_EXPIRATION']

    @abstractmethod
    def vendor_send_otp(self, phone, token) -> bool:
        pass

    def send_sms(self, phone):
        self.errors = None

        try:
            g = self.otp_generator
            token = g.gen_persist_otp(phone)
        except Exception:
            self.errors = "Failed to persist OTP to db"
            return False

        return self.vendor_send_otp(phone, token)

    def verify_sms(self, phone, token):
        g = self.otp_generator
        return g.verify_otp(phone, token)

    def get_errors(self):
        return self.errors



from typing import Callable
from flask import Blueprint
from flask_jwt_extended import JWTManager


class SecureAuth:
    """
    An object used to hold the context setting and callback functions for flask_securelogin
    """
    def __init__(self) -> None:
        """
        Create the SecureAuth instance. The instance contains
        - Blueprint object to dispatch Rest API
        - JWTManager object to manage access token and refresh token
        """
        self.bp = Blueprint('auth', __name__)
        self.jwt = JWTManager()
        self._create_sms_service_handler = None
        self._check_if_access_token_revoked_hanlder = None
        return

    def init_db(self, db) -> None:
        """
        Initialize the SecureAuth instance with database object. The DB object is used
        to insert/update/query DB for internal data models like User, ActiveToken

        :param db:
            An object of SQLAlchemy instance
        """
        self.db = db

        return

    def init_app(self, app) -> None:
        """
        Initialize the SecureAuth instance with a flask app object.

        :param app:
            An object of flask app
        """
        self.app = app

        # initialize jwt and bp
        self.jwt.init_app(app)
        app.register_blueprint(self.bp, url_prefix='/api')

        return

    def create_sms_service(self, callback: Callable) -> Callable:
        """
        This decorator sets callback function for creating sms service. The callback
        function will be used when it requests to create a service to send sms through
        external vendors.

        The decorated function must take 3 parameters.

        The first parameter is the db object that was initialized by init_db
        The second parameter is a dictionary containing configurations
        The third parameter is a string of phone number. The callback function can
        create different sms services per phone number country

        The decorated function must return an instance derived from SMSCall
        """
        self._create_sms_service_handler = callback
        return callback

    def get_create_sms_service_handler(self) -> Callable:
        """
        The function is to get the decorator function of create_sms_service.
        It is for internal use.
        """
        return self._create_sms_service_handler

    def check_if_access_token_revoked(self, callback: Callable) -> Callable:
        """
        This decorator sets callback function for checking whether access token is revoked.

        Usually an access token expires very soon. You don't have to implement this decorator.
        And implementation of checking access token will increase latency for every api
        request. However, if your application needs a better security and can tolerate the
        latency increase, you can implement this decorator. If you do so, it is strongly suggested
        that cache is used to store and check the access token.

        The decorated function must take 1 parameter.

        The first parameter is the jti of the access token.

        The decorated function must return Boolean. True means the token is revoked and invalid,
        false means token is not revoked and still valid.
        """
        self._check_if_access_token_revoked_hanlder = callback
        return callback

    def get_check_if_access_token_revoked_handler(self) -> Callable:
        """
        The function is to get the decorator function of check_if_access_token_revoked.
        It is for internal use.
        """
        return self._check_if_access_token_revoked_hanlder



from setuptools import setup
import setuptools

setup(
    name='flask-securelogin',
    version='0.1.8',
    description='A Flask-based installable library that provides secure login capabilities with SMS and auth tokens.',
    long_description=open('README.rst', 'r').read(),
    url='https://github.com/brilliance-qian/flask-securelogin',
    author='Brilliance Qian',
    license_files=('LICENSE.txt',),
    packages=setuptools.find_packages(),
    install_requires=[
        'aiohttp',
        'aiohttp-retry',
        'aiosignal',
        'alembic',
        'async-timeout',
        'attrs',
        'Babel',
        'bcrypt',
        'certifi',
        'charset-normalizer',
        'click',
        'exceptiongroup',
        'Flask',
        'flask-babel',
        'flask-blueprint',
        'Flask-JWT-Extended',
        'Flask-Migrate',
        'Flask-SQLAlchemy',
        'frozenlist',
        'greenlet',
        'idna',
        'importlib-metadata',
        'importlib-resources',
        'iniconfig',
        'itsdangerous',
        'Jinja2',
        'Mako',
        'MarkupSafe',
        'multidict',
        'packaging',
        'pluggy',
        'PyJWT',
        'pytest',
        'pytest-mock',
        'python-dateutil',
        'python-dotenv',
        'pytz',
        'requests',
        'six',
        'SQLAlchemy',
        'tencentcloud-sdk-python',
        'time-machine',
        'tomli',
        'twilio',
        'typing-extensions',
        'urllib3',
        'Werkzeug',
        'yarl',
        'zipp',
    ],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    extras_require={
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-clarity",
            'mock;python_version>="3.7"']
    }
)



from twilio.rest import Client
from flask_securelogin.sms.SMSCall import SMSCall


class TwilioClient(SMSCall):
    def __init__(self, db, config):
        client = Client(config['TWILIO_ACCOUNT_SID'], config['TWILIO_AUTH_TOKEN'])
        verify = client.verify.services(config['TWILIO_SMS_SID'])

        self.db = db
        self.client = client
        self.verify = verify
        self.errors = None

    def send_sms(self, phone):
        self.verify.verifications.create(to=phone, channel='sms')

        return True

    def verify_sms(self, phone, token):
        self.errors = None
        result = self.verify.verification_checks.create(to=phone, code=token)

        return result.status == 'approved'

    def get_errors(self):
        return self.errors



import flask
import logging
from flask_securelogin.exception import AppException
from flask_securelogin.errors import bad_request, error_response
from flask_securelogin.account import AccountManager
from flask_securelogin.tokens import TokenManager

from sqlalchemy import exc
from flask import request

from flask_jwt_extended import (
    get_jwt,
    jwt_required,
    get_jwt_identity
)

from flask_securelogin import secure_auth

bp = secure_auth.bp
jwt = secure_auth.jwt

log = logging.getLogger(__name__)


@bp.route('/auth/register', methods=['POST'])
def register():
    config = flask.current_app.config
    d = request.get_json()

    auth_type = d['auth_type']
    if auth_type not in ['PASSWORD', 'PHONE', 'APPLE_ID', 'GOOGLE_ID']:
        return bad_request("AccountCreationFailure", "unsupported authentication method")

    account_manager = AccountManager(secure_auth.db, config)
    new_user = None

    if auth_type == 'PASSWORD':
        username = d['username']
        email = d['email']
        password = d['password']

        new_user = account_manager.create_user_w_password(username, email, password)
    elif auth_type == 'PHONE':
        username = d['username']
        phone = d['phone']

        new_user = account_manager.create_user_w_phone(username, phone)

    if new_user is None:
        return bad_request("AccountCreationFailure", "account creation method is not supported yet")

    resp = {'message': 'registered successfully', 'userid': new_user.public_id}
    if auth_type == 'PASSWORD':
        token_manager = TokenManager(secure_auth.db)
        (access_token, refresh_token) = token_manager.start_new_session(new_user.id)
        resp['access_token'] = access_token
        resp['refresh_token'] = refresh_token

    return resp


@bp.route('/auth/login', methods=['POST'])
@bp.route('/auth/login_sms', methods=['POST'])
def login():
    d = request.get_json()

    auth_type = d['auth_type']

    if auth_type not in ['PASSWORD', 'PHONE', 'APPLE_ID', 'GOOGLE_ID']:
        return bad_request("LoginFailure", "unsupported authentication method")

    user = None
    config = flask.current_app.config
    account_manager = AccountManager(secure_auth.db, config)

    if auth_type == 'PASSWORD':
        email = d['email']
        password = d['password']

        user = account_manager.login_by_password(email, password)

        if user is None:
            return bad_request("LoginFailure", "unknown login failure")

        resp = {}
        resp['userid'] = user.public_id

        token_manager = TokenManager(secure_auth.db)
        (access_token, refresh_token) = token_manager.start_new_session(user.id)
        resp['access_token'] = access_token
        resp['refresh_token'] = refresh_token

        return resp
    elif auth_type == 'PHONE':
        phone = d['phone']

        user = account_manager.login_by_phone(phone)
        if user is None:
            return bad_request("LoginFailure", "unknown login failure")

        resp = {}
        resp['userid'] = user.public_id
        resp['phone'] = user.phone
        return resp
    elif auth_type == 'GOOGLE_ID':
        # TODO: implement later
        user = {}
        user['emai'] = d['email']
    elif auth_type == 'APPLE_ID':
        # TODO: implement later
        user = {}
        user['emai'] = d['email']

    return d


@bp.route('/auth/verify_sms', methods=['POST'])
def verify_sms():
    d = request.get_json()

    phone = d['phone']
    token = d['token']
    public_id = d['userid']

    config = flask.current_app.config
    account_manager = AccountManager(secure_auth.db, config)
    user = account_manager.login_verify_sms(phone, token)

    if user.public_id != public_id:
        raise AppException("LoginFailure", "SMSVerifyFailure", "invalid user account")

    resp = {}
    resp['userid'] = user.public_id

    token_manager = TokenManager(secure_auth.db)
    (access_token, refresh_token) = token_manager.start_new_session(user.id)

    resp['access_token'] = access_token
    resp['refresh_token'] = refresh_token

    return resp


@bp.route('/auth/logout', methods=['POST'])
@jwt_required()
def logout():
    d = request.get_json()

    refresh_token = d['refresh_token']
    access_token_jwt = get_jwt()

    token_manager = TokenManager(secure_auth.db)
    token_manager.logout_token(refresh_token, access_token_jwt)

    return {'message': 'logout successful'}


@bp.route('/auth/password', methods=['POST'])
@jwt_required(fresh=True)
def change_password():
    d = request.get_json()
    oldpassword = d['oldpassword']
    newpassword = d['newpassword']

    config = flask.current_app.config
    account_manager = AccountManager(secure_auth.db, config)

    userid = get_jwt_identity()
    account_manager.change_password(userid, oldpassword, newpassword)

    return {"message": 'changed password successfully'}


@bp.route('/auth/op', methods=['POST'])
@jwt_required()
def op():
    return {"message": 'test op successful'}


@bp.route('/auth/op2', methods=['POST'])
@jwt_required(fresh=True)
def op2():
    return {"message": 'test op2 successful'}


@bp.route('/auth/logout_all_other_sessions', methods=['POST'])
@jwt_required(fresh=True)
def logout_all_other_sessions():
    d = request.get_json()
    refresh_token = d['refresh_token']
    access_token_jwt = get_jwt()

    # logout all other sessions except for current one
    token_manager = TokenManager(secure_auth.db)
    token_manager.logout_all_other_sessions(access_token_jwt, refresh_token)

    return {'message': 'logout all other sessions successful'}


# We are using the `refresh=True` options in jwt_required to only allow
# refresh tokens to access this route.
@bp.route("/auth/refresh", methods=["POST"])
@jwt_required(refresh=True)
def refresh():
    token_manager = TokenManager(secure_auth.db)
    (access_token, refresh_token) = token_manager.refresh_tokens(get_jwt())

    resp = {}
    resp['access_token'] = access_token
    resp['refresh_token'] = refresh_token

    return resp


# Using the expired_token_loader decorator, we will now call
# this function whenever an expired but otherwise valid access
# token attempts to access an endpoint
@jwt.expired_token_loader
def expired_token_callback(jwt_header, jwt):
    if jwt['type'] == 'refresh':
        return error_response(401, "RefreshTokenExpired", message="The refresh_token is expired. Please login again")
    else:
        return error_response(401, "AccessTokenExpired",
                              message="The token is expired. Please refresh the access token")


# Using the needs_fresh_token_loader decorator, it will call
# this function when a fresh token is required but the token
# is not freshed anymore
@jwt.needs_fresh_token_loader
def token_not_fresh_callback(jwt_header, jwt_payload):
    return error_response(401, "FreshTokenRequired", message="The token is not fresh.")


# Check if the token is invalid.
# Token policy:
#   access token: either don't check and let it expired (expiration should be short)
#                 or check it against cache
#   refresh token: the expiration of refresh_token is relatively longer (30 or 60 days)
#                 it checks against DB whether the stoken is still active or not
@jwt.token_in_blocklist_loader
def check_if_token_revoked(jwt_header, jwt_payload: dict) -> bool:
    token_manager = TokenManager(secure_auth.db)
    return token_manager.check_if_token_revoked(jwt_payload)


@bp.app_errorhandler(AppException)
def handle_app_exception(e):
    log.warning("Exception caught: %s, %s" % (e.__class__.__name__, str(e)))
    return error_response(400, e.category, failure_reason=e.error, message=str(e))


@bp.app_errorhandler(exc.IntegrityError)
@bp.app_errorhandler(KeyError)
@bp.app_errorhandler(TypeError)
@bp.app_errorhandler(Exception)
def handle_exception(e):
    log.warning("Exception caught: %s, %s" % (e.__class__.__name__, str(e)))
    return bad_request(e.__class__.__name__, str(e))
